/**************************************************************************
 *
 * NAME
 *    encode.h
 *  
 * VERSION
 *    1.0
 *    
 * DATE
 *    10.9.1998
 *
 * IMPLEMENTATION 
 *    Bertram Ganz (ganz@informatik.uni-freiburg.de)
 *    Martin Mueller (mueller@www.fr.bw.schule.de)
 * 
 * MODIFICATIONS
 *   
 * DESCRIPTION
 *    Header for encode.c / encode.functions.c     
 *
 **************************************************************************/

/**************************************************************************/
/*                               Includes                                 */
/**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <unistd.h>

/**************************************************************************/
/*                               Definitions                             */
/**************************************************************************/

#define DEBUG 0
#define GREY_LEVELS 255
#define IMAGE_TYPE unsigned char /* may be different in some applications */

/* The following #define allocates an hsize x vsize  matrix of type TYPE */
#define matrix_allocate(matrix, hsize, vsize, TYPE) {\
    TYPE *imptr; \
    int _i; \
    matrix = (TYPE **)malloc((vsize)*sizeof(TYPE *));\
    imptr = (TYPE*)malloc((long)(hsize)*(long)(vsize)*sizeof(TYPE));\
    if (imptr == NULL) \
        fatal("\nNo memory in matrix allocate."); \
    for (_i = 0; _i<vsize; ++_i, imptr += hsize) \
        matrix[_i] = imptr; \
 }
#define swap(a,b,TYPE)           {TYPE _temp; _temp=b; b=a; a= _temp;}
#define bound(a)   ((a) < 0.0 ? 0 : ((a)>255.0? 255 : a))
#define START {if ((start=times(&tmsstart))==-1) printf("times error");}
#define STOP  {if ((end=times(&tmsend))==-1) printf("times error");}


/**************************************************************************/
/*                               DECLARATIONS                            ,*/
/**************************************************************************/

const double e[16];   /* Unit vector                                     */
const int rot_transform[2][8];
const int phi_trafo[8][16]; 
const int codelength[3][32];
const char huff_codeword[3][32][13];

 
IMAGE_TYPE **image;          /* The input image data                      */ 

double     **domimage[4];    /* Decimated input image used for domains    */
double max_scale,            /* Maximum allowable grey level scale factor */
       max_scale_vq[3];
double *list_dist, seconds;

int    s_bits,               /* Number of bits used to store scale factor */
       o_bits,               /* Number of bits used to store offset       */
       min_part,             /* Min and max _part determine a range of    */
       max_part,             /* Range sizes from hsize>>min to hsize>>max */
       dom_step,             /* Density of domains relative to size       */
       dom_step_type,        /* Flag for dom_step a multiplier or divisor */
       dom_type,             /* Method of generating domain pool 0,1,2..  */
       only_positive,        /* A flag specifying use of positive scaling */
       subclass_search,      /* A flag specifying classes searched        */
       fullclass_search,     /* A flag specifying classes searched        */
       *bits_needed,         /* Number of bits to encode domain position. */
       max_exponent,         /* The max power of 2 side of square image   */
                             /* that fits inour input image.              */
       dim,                  /* Dimension of  feature vectors             */
       best,                 /* number of clusters to be visited in near- */
                             /* est-neighbor search for best cl. center.  */
       number_nn_class,
       print_clu_bel,       
       c_x,                  /* x/y dimension of cluster centers          */
       c_y,
       offset_bits,          /* number of offset bits                     */
       *dim_clus,            /* number of non empty clusters for single   */
                             /* quadtree depth.                           */ 
       *nn,                  /* maximal number of classes to be visited   */
                             /* in nearest-neighbor search.               */
       size_clus,
       vec_bits,             /* number of bits we need to store the code- */
                             /* bookindex for VQ                          */ 
       *list_clu,
       *list_sgn,
       **list_perm;          /* list of permutations we get by a new      */
			     /* ordering for avoiding empty cells (cl.).  */

clock_t    start, end;

struct tms tmsstart, tmsend;

struct counter_data {
    int vq, fr, mean;
} counter;

struct domain_data {
   int *no_h_domains,    /* The number of domains horizontally for    */
       *no_v_domains,    /* each size.                                */
       *domain_hsize,    /* The size of the domain.                   */
       *domain_vsize,    /* The size of the domain.                   */
       *domain_hstep,    /* The density of the domains.               */
       *domain_vstep;    /* The density of the domains.               */
  struct domain_pixels { /* This is a three (sigh) index array that   */
     int dom_x, dom_y;   /* dynamically allocated. The first index is */
     double *phi_vec;    /* the domain size, the other are two its    */
     double sum, sum2;   /* position. Der 4. die shuffle-trafo        */
     int sym;            /* It contains the phi-vector                */
  } ***pixel;            /* of the pixel values in the domains, which */
} domain;                /* are computed just once.                   */


struct classified_domain {         /* This is a list which containes  */
   struct domain_pixels *the;      /* pointers to  the domain data    */
   struct classified_domain *next; /* in the structure above.         */
}***the_domain;                    /* c_x * c_y Cluster-centers!      */
                                   /* Using this array, only          */
                                   /* domains and ranges in the same  */
                                   /* class are compared..            */
                                   /* The first pointer points to the */
                                   /* domain size the the second to   */
                                   /* list of domains.                */
struct quadtree_node {
    int method,    /* fractal: 0, OMSGVQ: 1 */ 
        quadtree,  /* quadtree bit: down: 1, stop:0 */
        iso,       /* isometry operation */
        scal,      /* scaling factor */
        offset,    /* offset */
        block_x,   /* blocknumber: Domain(fractal) or clustercenter (VQ) */
        block_y,
        rate;        /* number of bits needed to store relevant data */
    double se;       /* summed square error */
    struct quadtree_node *son1, *son2, *son3, *son4;
};

struct coder_data {
    int method,    /* fractal: 0, OMRSG-VQ: 1 */ 
        quadtree,  /* Quadtree bit: down: 1, stop:0 */
        iso,       /* isometry operation */
        scal,      /* scaling factor */
        offset,    /* offset */
        block_x,   /* blocknumber: Domain(fractal) or cluster center (VQ) */
        block_y;
} ***code;       

double ***centers,    /* list of cluster centers    */
       **centers_64,  /* list of 64-dim  centers   */
       **centers_256; /* list of 256-dim centers   */

FILE *outputfile,
     *methodfile,
     *offsetfile;

/**************************************************************************/
/*                          function prototypes                           */
/**************************************************************************/

/* encode.c */
extern void   classify(int x, int y, int xsize, int ysize, int *pfirst, 
		       double **pphi, int *psym, double *psum, 
		       double *psum2, int type);
extern double compare_cen(int atx, int aty, int xsize, int ysize, int depth, 
			  double rsum, double rsum2,
			  int index,
			  int sym_op, int *pialpha, int *pibeta, 
			  double *alpha, double *beta);
extern double compare_fr(int atx, int aty, int xsize, int ysize, int depth, 
	      	         double rsum, double rsum2, int dom_x, int dom_y, 
		         int sym_op, int *pialpha, int *pibeta);
extern void   compute_sums(int hsize, int vsize);
extern void   encode_image(int hsize, int vsize, double lambda);
extern void   encode_quadtree(int atx, int aty, int blocksize, int depth,
			      double lambda, struct quadtree_node *root);


/* encode.functions.c */
extern void   average(int x, int y, int xsize, int ysize, double *psum,
		      double *psum2);
extern void   average1(int x, int y, int xsize, int ysize, double *psum, 
		       double *psum2);
extern void   cluster_search(double *phi, int *cl_nr, double *list, int *l_sgn,
			  signed int sgn, int last, int depth);
extern void   compute_phi(double *x, double **phi, int size, double psum);
extern double dist(double *a, double *b, signed int sgn);
extern double distnorm(double *a, double *b, signed int sgn);
extern void   fatal();
extern void   initialize();
extern void   list_free(struct classified_domain *node);
extern void   merge_outputdata(char *outputfilename, int hsize, int vsize);
extern double nn_cluster_search(double *phi, int *cl_nr, signed int sgn, 
				int depth);
extern double norm(double *a, int dim);  
extern int    pack(int size, long int value, int which_file, FILE *foutf);
extern void   prune_quadtree(double lambda, struct quadtree_node *node, 
			   int depth, int prune_depth);
extern void   quadtree_alloc(struct quadtree_node *node, int depth);
extern void   read_commandline(int *argc, char **argv,
		              double *tol, 
			      char *inputfilename, char *outputfilename, 
			      int *hsize, int *vsize, long *stripchar);
extern void   save_data(int atx, int aty, int blocksize, int depth);
extern void   save_data1(int atx, int aty, int blocksize, int depth);
extern void   store_qtdata(int atx, int aty, struct quadtree_node *node, 
		           int depth, int blocksize);
extern void   visit_leaves(struct quadtree_node *node, double *se_leaves,
			   int *rate_leaves, int *qt_bits, int depth);

